import { SourceNode } from '@journeyapps/source-utils';
import { VFS } from './VFS';
import { SourceEntry } from './SourceEntry';

export interface SingleTransformOptions {
  inputs: string[];
  target: string;
  isScript?: boolean;
  isModule?: boolean;
}

/**
 * Abstract class to simplify implementation of a VFS that generates a single
 * file from one or more inputs.
 */
export abstract class SingleTransform implements VFS {
  public readonly target: string;
  public readonly inputs: string[];
  public readonly isScript: boolean;
  public readonly isModule: boolean;

  protected constructor(options: SingleTransformOptions) {
    this.target = options.target;
    this.inputs = options.inputs;
    this.isScript = options.isScript ?? false;
    this.isModule = options.isModule ?? false;
  }

  abstract generateSingle(): SourceNode | null;
  abstract getVersion(): string | null;

  read(target: string): SourceNode | null {
    if (target == this.target) {
      return this.generateSingle();
    } else {
      return null;
    }
  }

  info(path: string): SourceEntry | null {
    if (path != this.target) {
      return null;
    }
    const version = this.getVersion();
    if (version == null) {
      return null;
    }
    return {
      path,
      version,
      generated: true,
      isScript: this.isScript,
      isModule: this.isModule
    };
  }

  listFiles(): string[] {
    return [this.target];
  }

  getTargetsFor(source: string): string[] | null {
    if (this.inputs.includes(source)) {
      return [this.target];
    }
    return null;
  }
}
